@@ -125,6 +125,11 @@ func (s *sqlLoader) addOperatorBundle(tx *sql.Tx, bundle *registry.Bundle) error
 		return err
 	}
 
+	err = s.addBundleProperties(tx, bundle)
+	if err != nil {
+		return err
+	}
+
 	return s.addAPIs(tx, bundle)
 }
 
@@ -300,6 +305,9 @@ func (s *sqlLoader) addPackageChannels(tx *sql.Tx, manifest registry.PackageMani
 	  SELECT DISTINCT operatorbundle.csv
 	  FROM operatorbundle
 	  WHERE operatorbundle.name=? LIMIT 1`)
+	if err != nil {
+		return err
+	}
 	defer getReplaces.Close()
 
 	var errs []error
@@ -346,12 +354,24 @@ func (s *sqlLoader) addPackageChannels(tx *sql.Tx, manifest registry.PackageMani
 		replaceCycle := map[string]bool{channelEntryCSVName: true}
 		for {
 			// Get CSV for current entry
-			replaces, skips, err := s.getBundleSkipsReplaces(tx, channelEntryCSVName)
+			replaces, skips, version, err := s.getBundleSkipsReplacesVersion(tx, channelEntryCSVName)
 			if err != nil {
 				errs = append(errs, err)
 				break
 			}
 
+			bundlePath, err := s.getBundlePathIfExists(tx, channelEntryCSVName)
+			if err != nil {
+				// this should only happen on an SQL error, bundlepath just not being set is for backwards compatibility reasons
+				errs = append(errs, err)
+				break
+			}
+
+			if err := s.addPackageProperty(tx, channelEntryCSVName, manifest.PackageName, version, bundlePath); err != nil {
+				errs = append(errs, err)
+				break
+			}
+
 			for _, skip := range skips {
 				// add dummy channel entry for the skipped version
 				skippedChannelEntry, err := addChannelEntry.Exec(c.Name, manifest.PackageName, skip, depth)
@@ -415,7 +435,7 @@ func (s *sqlLoader) addPackageChannels(tx *sql.Tx, manifest registry.PackageMani
 				errs = append(errs, err)
 				break
 			}
-			if _, _, err := s.getBundleSkipsReplaces(tx, replaces); err != nil {
+			if _, _, _, err := s.getBundleSkipsReplacesVersion(tx, replaces); err != nil {
 				errs = append(errs, fmt.Errorf("%s specifies replacement that couldn't be found", c.CurrentCSVName))
 				break
 			}
@@ -457,17 +477,17 @@ func (s *sqlLoader) ClearNonHeadBundles() error {
 	return tx.Commit()
 }
 
-func (s *sqlLoader) getBundleSkipsReplaces(tx *sql.Tx, bundleName string) (replaces string, skips []string, err error) {
-	getReplacesAndSkips, err := tx.Prepare(`
-	  SELECT replaces, skips
+func (s *sqlLoader) getBundleSkipsReplacesVersion(tx *sql.Tx, bundleName string) (replaces string, skips []string, version string, err error) {
+	getReplacesSkipsAndVersions, err := tx.Prepare(`
+	  SELECT replaces, skips, version
 	  FROM operatorbundle
 	  WHERE operatorbundle.name=? LIMIT 1`)
 	if err != nil {
 		return
 	}
-	defer getReplacesAndSkips.Close()
+	defer getReplacesSkipsAndVersions.Close()
 
-	rows, rerr := getReplacesAndSkips.Query(bundleName)
+	rows, rerr := getReplacesSkipsAndVersions.Query(bundleName)
 	if err != nil {
 		err = rerr
 		return
@@ -478,16 +498,52 @@ func (s *sqlLoader) getBundleSkipsReplaces(tx *sql.Tx, bundleName string) (repla
 	}
 
 	var replacesStringSQL sql.NullString
-	var skipsStringSql sql.NullString
-	if err = rows.Scan(&replacesStringSQL, &skipsStringSql); err != nil {
+	var skipsStringSQL sql.NullString
+	var versionStringSQL sql.NullString
+	if err = rows.Scan(&replacesStringSQL, &skipsStringSQL, &versionStringSQL); err != nil {
 		return
 	}
 
 	if replacesStringSQL.Valid {
 		replaces = replacesStringSQL.String
 	}
-	if skipsStringSql.Valid && len(skipsStringSql.String) > 0 {
-		skips = strings.Split(skipsStringSql.String, ",")
+	if skipsStringSQL.Valid && len(skipsStringSQL.String) > 0 {
+		skips = strings.Split(skipsStringSQL.String, ",")
+	}
+	if versionStringSQL.Valid {
+		version = versionStringSQL.String
+	}
+
+	return
+}
+
+func (s *sqlLoader) getBundlePathIfExists(tx *sql.Tx, bundleName string) (bundlePath string, err error) {
+	getBundlePath, err := tx.Prepare(`
+	  SELECT bundlepath
+	  FROM operatorbundle
+	  WHERE operatorbundle.name=? LIMIT 1`)
+	if err != nil {
+		return
+	}
+	defer getBundlePath.Close()
+
+	rows, rerr := getBundlePath.Query(bundleName)
+	if err != nil {
+		err = rerr
+		return
+	}
+	if !rows.Next() {
+		// no bundlepath set
+		return
+	}
+
+	var bundlePathSQL sql.NullString
+	if err = rows.Scan(&bundlePathSQL); err != nil {
+		return
+	}
+
+	if bundlePathSQL.Valid {
+		bundlePath = bundlePathSQL.String
 	}
 
 	return
@@ -705,13 +761,12 @@ func (s *sqlLoader) addDependencies(tx *sql.Tx, bundle *registry.Bundle) error {
 	}
 
 	for api := range requiredApis {
-		valueMap := map[string]string{
-			"type":    registry.GVKType,
-			"group":   api.Group,
-			"version": api.Version,
-			"kind":    api.Kind,
+		dep := registry.GVKDependency{
+			Group:   api.Group,
+			Kind:    api.Kind,
+			Version: api.Version,
 		}
-		value, err := json.Marshal(valueMap)
+		value, err := json.Marshal(dep)
 		if err != nil {
 			return err
 		}
@@ -722,3 +777,320 @@ func (s *sqlLoader) addDependencies(tx *sql.Tx, bundle *registry.Bundle) error {
 
 	return nil
 }
+
+func (s *sqlLoader) addProperty(tx *sql.Tx, propType, value, bundleName, version, path string) error {
+	addProp, err := tx.Prepare("insert into properties(type, value, operatorbundle_name, operatorbundle_version, operatorbundle_path) values(?, ?, ?, ?, ?)")
+	if err != nil {
+		return err
+	}
+	defer addProp.Close()
+
+	sqlString := func(s string) sql.NullString {
+		return sql.NullString{String: s, Valid: s != ""}
+	}
+
+	if _, err := addProp.Exec(propType, value, bundleName, sqlString(version), sqlString(path)); err != nil {
+		return err
+	}
+	return nil
+}
+
+func (s *sqlLoader) addPackageProperty(tx *sql.Tx, bundleName, pkg, version, bundlePath string) error {
+	// Add the package property
+	prop := registry.PackageProperty{
+		PackageName: pkg,
+		Version:     version,
+	}
+	value, err := json.Marshal(prop)
+	if err != nil {
+		return err
+	}
+
+	return s.addProperty(tx, registry.PackageType, string(value), bundleName, version, bundlePath)
+}
+
+func (s *sqlLoader) addBundleProperties(tx *sql.Tx, bundle *registry.Bundle) error {
+	bundleVersion, err := bundle.Version()
+	if err != nil {
+		return err
+	}
+
+	for _, prop := range bundle.Properties {
+		value, _ := json.Marshal(prop.Value)
+		if err := s.addProperty(tx, prop.Type, string(value), bundle.Name, bundleVersion, bundle.BundleImage); err != nil {
+			return err
+		}
+	}
+
+	// Look up providedAPIs in CSV and add them in properties table
+	providedApis, err := bundle.ProvidedAPIs()
+	if err != nil {
+		return err
+	}
+
+	for api := range providedApis {
+		prop := registry.GVKProperty{
+			Group:   api.Group,
+			Kind:    api.Kind,
+			Version: api.Version,
+		}
+		value, err := json.Marshal(prop)
+		if err != nil {
+			return err
+		}
+		if err := s.addProperty(tx, registry.GVKType, string(value), bundle.Name, bundleVersion, bundle.BundleImage); err != nil {
+			return err
+		}
+	}
+
+	// Add label properties
+	if csv, err := bundle.ClusterServiceVersion(); err == nil {
+		annotations := csv.ObjectMeta.GetAnnotations()
+		if v, ok := annotations[registry.PropertyKey]; ok {
+			var props []registry.Property
+			if err := json.Unmarshal([]byte(v), &props); err == nil {
+				for _, prop := range props {
+					// Only add label type from the list
+					// TODO: Support more types such as GVK and package
+					if prop.Type == registry.LabelType {
+						var label registry.LabelProperty
+						err := json.Unmarshal(prop.Value, &label)
+						if err != nil {
+							continue
+						}
+						value, err := json.Marshal(label)
+						if err != nil {
+							continue
+						}
+						if err := s.addProperty(tx, registry.LabelType, string(value), bundle.Name, bundleVersion, bundle.BundleImage); err != nil {
+							continue
+						}
+					}
+				}
+			}
+		}
+	}
+
+	return nil
+}
+
+func (s *sqlLoader) rmChannelEntry(tx *sql.Tx, csvName string) error {
+	getEntryID := `SELECT entry_id FROM channel_entry WHERE operatorbundle_name=?`
+	rows, err := tx.QueryContext(context.TODO(), getEntryID, csvName)
+	if err != nil {
+		return err
+	}
+	var entryIDs []int64
+	for rows.Next() {
+		var entryID sql.NullInt64
+		rows.Scan(&entryID)
+		entryIDs = append(entryIDs, entryID.Int64)
+	}
+	err = rows.Close()
+	if err != nil {
+		return err
+	}
+
+	updateChannelEntry, err := tx.Prepare(`UPDATE channel_entry SET replaces=NULL WHERE replaces=?`)
+	if err != nil {
+		return err
+	}
+	for _, id := range entryIDs {
+		if _, err := updateChannelEntry.Exec(id); err != nil {
+			return err
+		}
+	}
+	err = updateChannelEntry.Close()
+	if err != nil {
+		return err
+	}
+
+	stmt, err := tx.Prepare("DELETE FROM channel_entry WHERE operatorbundle_name=?")
+	if err != nil {
+		return err
+	}
+	defer stmt.Close()
+
+	if _, err := stmt.Exec(csvName); err != nil {
+		return err
+	}
+
+	return nil
+}
+
+func getTailFromBundle(tx *sql.Tx, name string) (bundles []string, err error) {
+	getReplacesSkips := `SELECT replaces, skips FROM operatorbundle WHERE name=?`
+	isDefaultChannelHead := `SELECT head_operatorbundle_name FROM channel 
+							INNER JOIN package ON channel.name = package.default_channel 
+							WHERE channel.head_operatorbundle_name = ?`
+
+	tail := make(map[string]struct{})
+	next := name
+
+	for next != "" {
+		rows, err := tx.QueryContext(context.TODO(), getReplacesSkips, next)
+		if err != nil {
+			return nil, err
+		}
+		var replaces sql.NullString
+		var skips sql.NullString
+		if rows.Next() {
+			if err := rows.Scan(&replaces, &skips); err != nil {
+				return nil, err
+			}
+		}
+		rows.Close()
+		if skips.Valid && skips.String != "" {
+			for _, skip := range strings.Split(skips.String, ",") {
+				tail[skip] = struct{}{}
+			}
+		}
+		if replaces.Valid && replaces.String != "" {
+			// check if replaces is the head of the defaultChannel
+			// if it is, the defaultChannel will be removed
+			// this is not allowed because we cannot know which channel to promote as the new default
+			rows, err := tx.QueryContext(context.TODO(), isDefaultChannelHead, replaces.String)
+			if err != nil {
+				return nil, err
+			}
+			if rows.Next() {
+				var defaultChannelHead sql.NullString
+				err := rows.Scan(&defaultChannelHead)
+				if err != nil {
+					return nil, err
+				}
+				if defaultChannelHead.Valid {
+					return nil, registry.ErrRemovingDefaultChannelDuringDeprecation
+				}
+			}
+			next = replaces.String
+			tail[replaces.String] = struct{}{}
+		} else {
+			next = ""
+		}
+	}
+	var allTails []string
+
+	for k := range tail {
+		allTails = append(allTails, k)
+	}
+
+	return allTails, nil
+
+}
+
+func getBundleNameAndVersionForImage(tx *sql.Tx, path string) (string, string, error) {
+	query := `SELECT name, version FROM operatorbundle WHERE bundlepath=? LIMIT 1`
+	rows, err := tx.QueryContext(context.TODO(), query, path)
+	if err != nil {
+		return "", "", err
+	}
+	defer rows.Close()
+
+	var name sql.NullString
+	var version sql.NullString
+	if rows.Next() {
+		if err := rows.Scan(&name, &version); err != nil {
+			return "", "", err
+		}
+	}
+	if name.Valid && version.Valid {
+		return name.String, version.String, nil
+	}
+	return "", "", registry.ErrBundleImageNotInDatabase
+}
+
+func (s *sqlLoader) DeprecateBundle(path string) error {
+	tx, err := s.db.Begin()
+	if err != nil {
+		return err
+	}
+	defer func() {
+		tx.Rollback()
+	}()
+
+	name, version, err := getBundleNameAndVersionForImage(tx, path)
+	if err != nil {
+		return err
+	}
+	tailBundles, err := getTailFromBundle(tx, name)
+	if err != nil {
+		return err
+	}
+
+	for _, bundle := range tailBundles {
+		err := s.rmBundle(tx, bundle)
+		if err != nil {
+			return err
+		}
+		err = s.rmChannelEntry(tx, bundle)
+		if err != nil {
+			return err
+		}
+	}
+
+	deprecatedValue, err := json.Marshal(registry.DeprecatedProperty{})
+	if err != nil {
+		return err
+	}
+	err = s.addProperty(tx, registry.DeprecatedType, string(deprecatedValue), name, version, path)
+	if err != nil {
+		return err
+	}
+
+	return tx.Commit()
+}
+
+func (s *sqlLoader) RemoveStrandedBundles() ([]string, error) {
+	tx, err := s.db.Begin()
+	if err != nil {
+		return nil, err
+	}
+	defer func() {
+		tx.Rollback()
+	}()
+
+	bundles, err := s.rmStrandedBundles(tx)
+	if err != nil {
+		return nil, err
+	}
+
+	return bundles, tx.Commit()
+}
+
+func (s *sqlLoader) rmStrandedBundles(tx *sql.Tx) ([]string, error) {
+	strandedBundles := make([]string, 0)
+
+	strandedBundleQuery := `SELECT name FROM operatorbundle WHERE name NOT IN (select operatorbundle_name from channel_entry)`
+	rows, err := tx.QueryContext(context.TODO(), strandedBundleQuery)
+	if err != nil {
+		return nil, err
+	}
+
+	var name sql.NullString
+	for rows.Next() {
+		if err := rows.Scan(&name); err != nil {
+			return nil, err
+		}
+		if name.Valid {
+			strandedBundles = append(strandedBundles, fmt.Sprintf(`"%s"`, name.String))
+		}
+	}
+	rows.Close()
+
+	if len(strandedBundles) == 0 {
+		return nil, nil
+	}
+
+	rmStmt, err := tx.Prepare(fmt.Sprintf("DELETE FROM operatorbundle WHERE name IN(%s)", strings.Join(strandedBundles, ",")))
+	if err != nil {
+		return nil, err
+	}
+	defer rmStmt.Close()
+
+	if _, err := rmStmt.Exec(); err != nil {
+		return strandedBundles, err
+	}
+
+	return strandedBundles, nil
+}
