@@ -5,6 +5,7 @@ package sqlite
 import (
 	"context"
 	"database/sql"
+	"encoding/json"
 	"fmt"
 	"strings"
 
@@ -256,9 +257,14 @@ func (s *SQLQuerier) GetBundle(ctx context.Context, pkgName, channelName, csvNam
 	if err != nil {
 		return nil, err
 	}
-
 	out.Dependencies = dependencies
 
+	properties, err := s.GetPropertiesForBundle(ctx, name.String, version.String, bundlePath.String)
+	if err != nil {
+		return nil, err
+	}
+	out.Properties = properties
+
 	return out, nil
 }
 
@@ -311,9 +317,14 @@ func (s *SQLQuerier) GetBundleForChannel(ctx context.Context, pkgName string, ch
 	if err != nil {
 		return nil, err
 	}
-
 	out.Dependencies = dependencies
 
+	properties, err := s.GetPropertiesForBundle(ctx, name.String, version.String, bundlePath.String)
+	if err != nil {
+		return nil, err
+	}
+	out.Properties = properties
+
 	return out, nil
 }
 
@@ -402,22 +413,36 @@ func (s *SQLQuerier) GetBundleThatReplaces(ctx context.Context, name, pkgName, c
 	if err != nil {
 		return nil, err
 	}
-
 	out.Dependencies = dependencies
 
+	properties, err := s.GetPropertiesForBundle(ctx, outName.String, version.String, bundlePath.String)
+	if err != nil {
+		return nil, err
+	}
+	out.Properties = properties
+
 	return out, nil
 }
 
 func (s *SQLQuerier) GetChannelEntriesThatProvide(ctx context.Context, group, version, kind string) (entries []*registry.ChannelEntry, err error) {
+	// TODO: join on full fk, not just operatorbundlename
 	query := `SELECT DISTINCT channel_entry.package_name, channel_entry.channel_name, channel_entry.operatorbundle_name, replaces.operatorbundle_name
           FROM channel_entry
-          INNER JOIN api_provider ON channel_entry.operatorbundle_name = api_provider.operatorbundle_name
+          INNER JOIN properties ON channel_entry.operatorbundle_name = properties.operatorbundle_name
           LEFT OUTER JOIN channel_entry replaces ON channel_entry.replaces = replaces.entry_id
-		  WHERE api_provider.group_name = ? AND api_provider.version = ? AND api_provider.kind = ?`
+		  WHERE properties.type=? AND properties.value=?`
 
-	rows, err := s.db.QueryContext(ctx, query, group, version, kind)
+	value, err := json.Marshal(map[string]string{
+		"group":   group,
+		"version": version,
+		"kind":    kind,
+	})
 	if err != nil {
-		return
+		return nil, err
+	}
+	rows, err := s.db.QueryContext(ctx, query, registry.GVKType, string(value))
+	if err != nil {
+		return nil, err
 	}
 	defer rows.Close()
 
@@ -450,11 +475,21 @@ func (s *SQLQuerier) GetChannelEntriesThatProvide(ctx context.Context, group, ve
 func (s *SQLQuerier) GetLatestChannelEntriesThatProvide(ctx context.Context, group, version, kind string) (entries []*registry.ChannelEntry, err error) {
 	query := `SELECT DISTINCT channel_entry.package_name, channel_entry.channel_name, channel_entry.operatorbundle_name, replaces.operatorbundle_name, MIN(channel_entry.depth)
           FROM channel_entry
-          INNER JOIN api_provider ON channel_entry.operatorbundle_name = api_provider.operatorbundle_name
+          INNER JOIN properties ON channel_entry.operatorbundle_name = properties.operatorbundle_name
 		  LEFT OUTER JOIN channel_entry replaces ON channel_entry.replaces = replaces.entry_id
-		  WHERE api_provider.group_name = ? AND api_provider.version = ? AND api_provider.kind = ?
+		  WHERE properties.type = ? AND properties.value = ?
 		  GROUP BY channel_entry.package_name, channel_entry.channel_name`
-	rows, err := s.db.QueryContext(ctx, query, group, version, kind)
+
+	value, err := json.Marshal(map[string]string{
+		"group":   group,
+		"version": version,
+		"kind":    kind,
+	})
+	if err != nil {
+		return nil, err
+	}
+
+	rows, err := s.db.QueryContext(ctx, query, registry.GVKType, string(value))
 	if err != nil {
 		return nil, err
 	}
@@ -491,12 +526,17 @@ func (s *SQLQuerier) GetBundleThatProvides(ctx context.Context, group, apiVersio
 	query := `SELECT DISTINCT channel_entry.entry_id, operatorbundle.bundle, operatorbundle.bundlepath, MIN(channel_entry.depth), channel_entry.operatorbundle_name, channel_entry.package_name, channel_entry.channel_name, channel_entry.replaces, operatorbundle.version, operatorbundle.skiprange
           FROM channel_entry
 		  INNER JOIN operatorbundle ON operatorbundle.name = channel_entry.operatorbundle_name
-		  INNER JOIN api_provider ON channel_entry.operatorbundle_name = api_provider.operatorbundle_name
+		  INNER JOIN properties ON channel_entry.operatorbundle_name = properties.operatorbundle_name
 		  INNER JOIN package ON package.name = channel_entry.package_name
-		  WHERE api_provider.group_name = ? AND api_provider.version = ? AND api_provider.kind = ? AND package.default_channel = channel_entry.channel_name
+		  WHERE properties.type = ? AND properties.value = ? AND package.default_channel = channel_entry.channel_name
 		  GROUP BY channel_entry.package_name, channel_entry.channel_name`
 
-	rows, err := s.db.QueryContext(ctx, query, group, apiVersion, kind)
+	value, err := json.Marshal(map[string]string{
+		"group":   group,
+		"version": apiVersion,
+		"kind":    kind,
+	})
+	rows, err := s.db.QueryContext(ctx, query, registry.GVKType, string(value))
 	if err != nil {
 		return nil, err
 	}
@@ -548,9 +588,14 @@ func (s *SQLQuerier) GetBundleThatProvides(ctx context.Context, group, apiVersio
 	if err != nil {
 		return nil, err
 	}
-
 	out.Dependencies = dependencies
 
+	properties, err := s.GetPropertiesForBundle(ctx, bundleName.String, version.String, bundlePath.String)
+	if err != nil {
+		return nil, err
+	}
+	out.Properties = properties
+
 	return out, nil
 }
 
@@ -596,75 +641,148 @@ func (s *SQLQuerier) GetImagesForBundle(ctx context.Context, csvName string) ([]
 }
 
 func (s *SQLQuerier) GetApisForEntry(ctx context.Context, entryID int64) (provided []*api.GroupVersionKind, required []*api.GroupVersionKind, err error) {
-	providedQuery := `SELECT DISTINCT api.group_name, api.version, api.kind, api.plural FROM api
-		 	  		  INNER JOIN channel_entry ON channel_entry.operatorbundle_name = api_provider.operatorbundle_name
-					  INNER JOIN operatorbundle ON operatorbundle.name=channel_entry.operatorbundle_name
-		 	  		  INNER JOIN api_provider ON (api.group_name=api_provider.group_name AND api.version=api_provider.version AND api.kind=api_provider.kind AND operatorbundle.name=api_provider.operatorbundle_name)
-			  		  WHERE channel_entry.entry_id=?`
+	groups := map[string]struct{}{}
+	kinds := map[string]struct{}{}
+	versions := map[string]struct{}{}
 
-	providedRows, err := s.db.QueryContext(ctx, providedQuery, entryID)
+	providedQuery := `SELECT properties.value FROM properties 
+		 	  		  INNER JOIN channel_entry ON channel_entry.operatorbundle_name = properties.operatorbundle_name
+			  		  WHERE properties.type=? AND channel_entry.entry_id=?`
+
+	providedRows, err := s.db.QueryContext(ctx, providedQuery, registry.GVKType, entryID)
 	if err != nil {
 		return nil, nil, err
 	}
+	defer providedRows.Close()
 
 	provided = []*api.GroupVersionKind{}
 	for providedRows.Next() {
-		var groupName sql.NullString
-		var versionName sql.NullString
-		var kindName sql.NullString
-		var pluralName sql.NullString
+		var value sql.NullString
 
-		if err := providedRows.Scan(&groupName, &versionName, &kindName, &pluralName); err != nil {
+		if err := providedRows.Scan(&value); err != nil {
 			return nil, nil, err
 		}
-		if !groupName.Valid || !versionName.Valid || !kindName.Valid || !pluralName.Valid {
-			return nil, nil, err
+
+		if !value.Valid {
+			continue
 		}
+		prop := registry.GVKProperty{}
+		if err := json.Unmarshal([]byte(value.String), &prop); err != nil {
+			continue
+		}
+
 		provided = append(provided, &api.GroupVersionKind{
-			Group:   groupName.String,
-			Version: versionName.String,
-			Kind:    kindName.String,
-			Plural:  pluralName.String,
+			Group:   prop.Group,
+			Version: prop.Version,
+			Kind:    prop.Kind,
 		})
-	}
-	if err := providedRows.Close(); err != nil {
-		return nil, nil, err
+		groups[prop.Group] = struct{}{}
+		versions[prop.Version] = struct{}{}
+		kinds[prop.Kind] = struct{}{}
 	}
 
-	requiredQuery := `SELECT DISTINCT api.group_name, api.version, api.kind, api.plural FROM api
-		 	  		  INNER JOIN channel_entry ON channel_entry.operatorbundle_name = api_requirer.operatorbundle_name
-		 	  		  INNER JOIN operatorbundle ON operatorbundle.name=channel_entry.operatorbundle_name
-		 	  		  INNER JOIN api_requirer ON (api.group_name=api_requirer.group_name AND api.version=api_requirer.version AND api.kind=api_requirer.kind AND operatorbundle.name=api_requirer.operatorbundle_name)
-			  		  WHERE channel_entry.entry_id=?`
+	requiredQuery := `SELECT DISTINCT dependencies.value FROM dependencies
+					  INNER JOIN channel_entry ON channel_entry.operatorbundle_name = dependencies.operatorbundle_name
+					  WHERE dependencies.type=? AND channel_entry.entry_id=?`
 
-	requiredRows, err := s.db.QueryContext(ctx, requiredQuery, entryID)
+	requiredRows, err := s.db.QueryContext(ctx, requiredQuery, registry.GVKType, entryID)
 	if err != nil {
 		return nil, nil, err
 	}
+	defer requiredRows.Close()
+
 	required = []*api.GroupVersionKind{}
 	for requiredRows.Next() {
+		var value sql.NullString
+
+		if err := requiredRows.Scan(&value); err != nil {
+			return nil, nil, err
+		}
+		if !value.Valid {
+			continue
+		}
+		dep := registry.GVKDependency{}
+		if err := json.Unmarshal([]byte(value.String), &dep); err != nil {
+			continue
+		}
+
+		required = append(required, &api.GroupVersionKind{
+			Group:   dep.Group,
+			Version: dep.Version,
+			Kind:    dep.Kind,
+		})
+		groups[dep.Group] = struct{}{}
+		versions[dep.Version] = struct{}{}
+		kinds[dep.Kind] = struct{}{}
+	}
+
+	argsFor := func(s map[string]struct{}) string {
+		l := []string{}
+		for v := range s {
+			l = append(l, "\""+v+"\"")
+		}
+		return "(" + strings.Join(l, ",") + ")"
+	}
+
+	pluralQuery := `SELECT * FROM api` +
+		` WHERE api.group_name IN ` + argsFor(groups) +
+		` AND api.version IN ` + argsFor(versions) +
+		` AND api.kind IN ` + argsFor(kinds)
+
+	pluralRows, err := s.db.QueryContext(ctx, pluralQuery)
+	if err != nil {
+		return nil, nil, err
+	}
+	defer pluralRows.Close()
+
+	gvkToPlural := map[registry.GVKProperty]string{}
+	for pluralRows.Next() {
 		var groupName sql.NullString
 		var versionName sql.NullString
 		var kindName sql.NullString
 		var pluralName sql.NullString
 
-		if err := requiredRows.Scan(&groupName, &versionName, &kindName, &pluralName); err != nil {
-			return nil, nil, err
+		if err := pluralRows.Scan(&groupName, &versionName, &kindName, &pluralName); err != nil {
+			continue
 		}
 		if !groupName.Valid || !versionName.Valid || !kindName.Valid || !pluralName.Valid {
-			return nil, nil, err
+			continue
 		}
-		required = append(required, &api.GroupVersionKind{
+		gvkToPlural[registry.GVKProperty{
 			Group:   groupName.String,
 			Version: versionName.String,
 			Kind:    kindName.String,
-			Plural:  pluralName.String,
-		})
-	}
-	if err := requiredRows.Close(); err != nil {
-		return nil, nil, err
+		}] = pluralName.String
 	}
 
+	for i, p := range provided {
+		if p.Plural != "" {
+			continue
+		}
+		plural, ok := gvkToPlural[registry.GVKProperty{
+			Group:   p.Group,
+			Version: p.Version,
+			Kind:    p.Kind,
+		}]
+		if !ok {
+			continue
+		}
+		provided[i].Plural = plural
+	}
+	for i, r := range required {
+		if r.Plural != "" {
+			continue
+		}
+		plural, ok := gvkToPlural[registry.GVKProperty{
+			Group:   r.Group,
+			Version: r.Version,
+			Kind:    r.Kind,
+		}]
+		if !ok {
+			continue
+		}
+		required[i].Plural = plural
+	}
 	return
 }
 
@@ -685,7 +803,7 @@ func (s *SQLQuerier) GetBundleVersion(ctx context.Context, image string) (string
 	if version.Valid {
 		return version.String, nil
 	}
-	return "", fmt.Errorf("bundle %s not found", image)
+	return "", nil
 }
 
 func (s *SQLQuerier) GetBundlePathsForPackage(ctx context.Context, pkgName string) ([]string, error) {
@@ -866,7 +984,9 @@ SELECT
     operatorbundle.version,
     operatorbundle.skiprange,
     dependencies.type,
-    dependencies.value
+    dependencies.value,
+    properties.type,
+    properties.value
   FROM replaces_bundle
     INNER JOIN operatorbundle
       ON replaces_bundle.operatorbundle_name = operatorbundle.name
@@ -875,7 +995,9 @@ SELECT
         AND replaces_bundle.package_name = skips_bundle.package_name
         AND replaces_bundle.channel_name = skips_bundle.channel_name
     LEFT OUTER JOIN dependencies
-      ON operatorbundle.name = dependencies.operatorbundle_name`
+      ON operatorbundle.name = dependencies.operatorbundle_name
+    LEFT OUTER JOIN properties
+      ON operatorbundle.name = properties.operatorbundle_name`
 
 func (s *SQLQuerier) ListBundles(ctx context.Context) ([]*api.Bundle, error) {
 	rows, err := s.db.QueryContext(ctx, listBundlesQuery)
@@ -900,8 +1022,10 @@ func (s *SQLQuerier) ListBundles(ctx context.Context) ([]*api.Bundle, error) {
 			skipRange   sql.NullString
 			depType     sql.NullString
 			depValue    sql.NullString
+			propType    sql.NullString
+			propValue   sql.NullString
 		)
-		if err := rows.Scan(&entryID, &bundle, &bundlePath, &bundleName, &pkgName, &channelName, &replaces, &skips, &version, &skipRange, &depType, &depValue); err != nil {
+		if err := rows.Scan(&entryID, &bundle, &bundlePath, &bundleName, &pkgName, &channelName, &replaces, &skips, &version, &skipRange, &depType, &depValue, &propType, &propValue); err != nil {
 			return nil, err
 		}
 
@@ -918,6 +1042,13 @@ func (s *SQLQuerier) ListBundles(ctx context.Context) ([]*api.Bundle, error) {
 					Value: depValue.String,
 				})
 			}
+
+			if propType.Valid && propValue.Valid {
+				bundleItem.Properties = append(bundleItem.Properties, &api.Property{
+					Type:  propType.String,
+					Value: propValue.String,
+				})
+			}
 		} else {
 			// Create new bundle
 			out := &api.Bundle{}
@@ -957,6 +1088,13 @@ func (s *SQLQuerier) ListBundles(ctx context.Context) ([]*api.Bundle, error) {
 				}}
 			}
 
+			if propType.Valid && propValue.Valid {
+				out.Properties = []*api.Property{{
+					Type:  propType.String,
+					Value: propValue.String,
+				}}
+			}
+
 			bundlesMap[bundleKey] = out
 		}
 	}
@@ -966,6 +1104,10 @@ func (s *SQLQuerier) ListBundles(ctx context.Context) ([]*api.Bundle, error) {
 			newDeps := unique(v.Dependencies)
 			v.Dependencies = newDeps
 		}
+		if len(v.Properties) > 1 {
+			newProps := uniqueProps(v.Properties)
+			v.Properties = newProps
+		}
 		bundles = append(bundles, v)
 	}
 
@@ -985,16 +1127,31 @@ func unique(deps []*api.Dependency) []*api.Dependency {
 	return list
 }
 
+func uniqueProps(props []*api.Property) []*api.Property {
+	keys := make(map[string]struct{})
+	var list []*api.Property
+	for _, entry := range props {
+		propKey := fmt.Sprintf("%s/%s", entry.Type, entry.Value)
+		if _, value := keys[propKey]; !value {
+			keys[propKey] = struct{}{}
+			list = append(list, entry)
+		}
+	}
+	return list
+}
+
 func (s *SQLQuerier) GetDependenciesForBundle(ctx context.Context, name, version, path string) (dependencies []*api.Dependency, err error) {
-	depQuery := `SELECT DISTINCT dependencies.type, dependencies.value FROM dependencies
-	WHERE dependencies.operatorbundle_name=?
-	AND (dependencies.operatorbundle_version=? OR dependencies.operatorbundle_version is NULL)
-	AND (dependencies.operatorbundle_path=? OR dependencies.operatorbundle_path is NULL)`
+	depQuery := `SELECT DISTINCT type, value FROM dependencies
+	WHERE operatorbundle_name=?
+	AND (operatorbundle_version=? OR operatorbundle_version is NULL)
+	AND (operatorbundle_path=? OR operatorbundle_path is NULL)`
 
 	rows, err := s.db.QueryContext(ctx, depQuery, name, version, path)
 	if err != nil {
 		return nil, err
 	}
+	defer rows.Close()
+
 	dependencies = []*api.Dependency{}
 	for rows.Next() {
 		var typeName sql.NullString
@@ -1011,9 +1168,66 @@ func (s *SQLQuerier) GetDependenciesForBundle(ctx context.Context, name, version
 			Value: value.String,
 		})
 	}
-	if err := rows.Close(); err != nil {
+
+	return
+}
+
+func (s *SQLQuerier) GetPropertiesForBundle(ctx context.Context, name, version, path string) (properties []*api.Property, err error) {
+	propQuery := `SELECT DISTINCT type, value FROM properties 
+				 WHERE operatorbundle_name=?
+				 AND (operatorbundle_version=? OR operatorbundle_version is NULL)
+				 AND (operatorbundle_path=? OR operatorbundle_path is NULL)`
+
+	rows, err := s.db.QueryContext(ctx, propQuery, name, version, path)
+	if err != nil {
 		return nil, err
 	}
+	defer rows.Close()
+
+	properties = []*api.Property{}
+	for rows.Next() {
+		var typeName sql.NullString
+		var value sql.NullString
+
+		if err := rows.Scan(&typeName, &value); err != nil {
+			return nil, err
+		}
+		if !typeName.Valid || !value.Valid {
+			return nil, err
+		}
+		properties = append(properties, &api.Property{
+			Type:  typeName.String,
+			Value: value.String,
+		})
+	}
+
+	return
+}
+
+func (s *SQLQuerier) GetBundlePathIfExists(ctx context.Context, bundleName string) (bundlePath string, err error) {
+	getBundlePathQuery := `
+	  SELECT bundlepath
+	  FROM operatorbundle
+	  WHERE operatorbundle.name=? LIMIT 1`
+
+	rows, err := s.db.QueryContext(ctx, getBundlePathQuery, bundleName)
+	if err != nil {
+		return
+	}
+	if !rows.Next() {
+		// no bundlepath set
+		err = registry.ErrBundleImageNotInDatabase
+		return
+	}
+
+	var bundlePathSQL sql.NullString
+	if err = rows.Scan(&bundlePathSQL); err != nil {
+		return
+	}
+
+	if bundlePathSQL.Valid {
+		bundlePath = bundlePathSQL.String
+	}
 
 	return
 }
